"""Run a process and tee its output to an email message and the console."""

from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Annotated

import typer

from ansi2html.converter import Ansi2HTMLConverter
from typer.core import TyperGroup

from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.DoneManager import DoneManager, Flags as DoneManagerFlags
from dbrownell_Common.Streams.StreamDecorator import StreamDecorator
from dbrownell_Common import SubprocessEx
from dbrownell_Email.SmtpMailer import SmtpMailer


# ----------------------------------------------------------------------
class NaturalOrderGrouper(TyperGroup):  # noqa: D101
    # ----------------------------------------------------------------------
    def list_commands(self, *args, **kwargs) -> list[str]:  # noqa: ARG002, D102
        return list(self.commands.keys())


# ----------------------------------------------------------------------
app = typer.Typer(
    cls=NaturalOrderGrouper,
    help=__doc__,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


# ----------------------------------------------------------------------
@app.command(
    "EntryPoint",
    help=__doc__,
    no_args_is_help=True,
)
def EntryPoint(  # noqa: D103
    command_line: Annotated[
        str,
        typer.Argument(
            ...,
            help="Command line to invoke; all output will be included in the email message. If the argument begins with '@', the rest of the command line will be interpreted as a filename and the command line will be read from that file.",
        ),
    ],
    smtp_profile_name: Annotated[
        str,
        typer.Argument(..., help="SMTP profile name."),
    ],
    email_recipients: Annotated[
        list[str],
        typer.Argument(..., help="Recipient(s) for the email message."),
    ],
    email_subject: Annotated[
        str,
        typer.Argument(
            ...,
            help="Subject of the email message; '{now}' can be used in the string as a template placeholder for the current time.",
        ),
    ],
    output_filename: Annotated[
        Path | None,
        typer.Option(
            "--output-filename",
            dir_okay=False,
            resolve_path=True,
            help="Writes formatted html output to a file; this is useful when --force-color has also been specified as an argument.",
        ),
    ] = None,
    background_color: Annotated[
        str,
        typer.Option("--background-color", help="Email background color."),
    ] = "black",
    verbose: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--verbose", help="Write verbose information to the terminal."),
    ] = False,
    debug: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--debug", help="Write debug information to the terminal."),
    ] = False,
) -> None:
    with DoneManager.CreateCommandLine(
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        try:
            smtp_mailer = SmtpMailer.Load(smtp_profile_name)
        except Exception as ex:
            dm.WriteError(str(ex))
            return

        with dm.Nested(
            "Running command...",
            suffix="\n",
        ) as running_dm:
            # Create the stream used to capture the message content
            message_sink = StringIO()

            Capabilities.Set(
                message_sink,
                Capabilities(
                    is_interactive=False,
                    supports_colors=True,
                    is_headless=True,
                    no_column_warning=True,
                ),
            )

            with running_dm.YieldStream() as dm_stream:
                running_dm.result = SubprocessEx.Stream(
                    command_line,
                    StreamDecorator([message_sink, dm_stream]),
                )

            message = message_sink.getvalue()

        with dm.Nested(
            "Processing output...",
            suffix="\n",
        ) as processing_dm:
            title = None

            if output_filename:
                title = output_filename.stem

            # Value to convert spaces into before the text is converted to html.
            space_placeholder = "__nbsp;__"

            with processing_dm.Nested("Converting output to HTML..."):
                message = message.replace(" ", space_placeholder)

                message = Ansi2HTMLConverter(
                    dark_bg=True,
                    inline=True,
                    line_wrap=False,
                    title=title or "",
                ).convert(message)

            for source, dest in [
                (space_placeholder, "&nbsp;"),
                # Create a div to set the background color
                (
                    '<pre class="ansi2html-content">\n',
                    '<pre class="ansi2html-content">\n<div style="background-color: {}">\n'.format(
                        background_color
                    ),
                ),
                # Undo the div that set the background color
                (
                    "</pre>\n",
                    "</div>\n</pre>\n",
                ),
            ]:
                message = message.replace(source, dest)

            if output_filename is not None:
                with processing_dm.Nested("Writing to '{}'...".format(output_filename)):
                    output_filename.parent.mkdir(parents=True, exist_ok=True)

                    with output_filename.open("w", encoding="utf-8") as f:
                        f.write(message)

        with dm.Nested("Sending email...") as email_dm:
            try:
                smtp_mailer.SendMessage(
                    email_recipients,
                    email_subject.format(now=datetime.now()),  # noqa: DTZ005
                    message,
                    message_format="html",
                )
            except Exception as ex:
                email_dm.WriteError(str(ex))
                return


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
if __name__ == "__main__":
    app()



"""Contains the SmtpMailer object."""

import datetime
import json
import mimetypes
import os
import smtplib
import ssl
import textwrap

from dataclasses import dataclass, field
from email import encoders
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from collections.abc import Generator
from pathlib import Path

from dbrownell_Common.ContextlibEx import ExitStack


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class SmtpMailer:
    """Code that manages SMTP profiles and uses them to send messages."""

    # ----------------------------------------------------------------------
    # |  Public Types
    PROFILE_EXTENSION = ".SmtpMailer"

    # ----------------------------------------------------------------------
    # |  Public Data
    host: str
    username: str
    password: str
    from_name: str
    from_email: str

    ssl: bool = field(kw_only=True)

    port: int | None = field(default=None)

    # ----------------------------------------------------------------------
    # |  Public Methods
    def ToString(
        self,
        *,
        show_password: bool = False,
    ) -> str:
        """Return a string representation of the profile."""

        return textwrap.dedent(
            """\
            host        : {host}
            username    : {username}
            password    : {password}
            from_name   : {from_name}
            from_email  : {from_email}
            ssl         : {ssl}
            port        : {port}
            """,
        ).format(
            host=self.host,
            username=self.username,
            password=self.password if show_password else "****",
            from_name=self.from_name,
            from_email=self.from_email,
            ssl=self.ssl,
            port=self.port,
        )

    # ----------------------------------------------------------------------
    def Save(
        self,
        profile_name: str,
    ) -> None:
        """Save a profile."""

        content = json.dumps(self.__dict__).encode("utf-8")

        if os.name == "nt":
            import win32crypt

            content = win32crypt.CryptProtectData(content, "", None, None, None, 0)

        with (Path("~").expanduser() / (profile_name + self.__class__.PROFILE_EXTENSION)).open("wb") as f:
            f.write(content)

    # ----------------------------------------------------------------------
    def SendMessage(
        self,
        recipients: list[str],
        subject: str,
        message: str,
        attachment_filenames: list[Path] | None = None,
        message_format: str = "plain",  # "html"
    ) -> None:
        """Send an email message using the current profile."""

        if self.ssl:
            port = self.port or 465
            smtp = smtplib.SMTP_SSL(self.host, port, context=ssl.create_default_context())
        else:
            port = self.port or 26
            smtp = smtplib.SMTP(self.host, port)

        smtp.connect(self.host, port)
        with ExitStack(smtp.close):
            if not self.ssl:
                smtp.starttls()

            smtp.login(self.username, self.password)

            from_addr = "{} <{}>".format(self.from_name, self.from_email)

            msg = MIMEMultipart() if attachment_filenames else MIMEMultipart("alternative")

            msg["Subject"] = subject
            msg["From"] = from_addr
            msg["To"] = ", ".join(recipients)

            msg.attach(MIMEText(message, message_format))

            for attachment_filename in attachment_filenames or []:
                ctype, encoding = mimetypes.guess_type(attachment_filename)

                if ctype is None or encoding is not None:
                    ctype = "application/octet-stream"

                maintype, subtype = ctype.split("/", 1)

                with attachment_filename.open("rb") as f:
                    content = f.read()

                if maintype == "text":
                    attachment = MIMEText(content.decode("utf-8"), _subtype=subtype)
                elif maintype == "image":
                    attachment = MIMEImage(content, _subtype=subtype)
                elif maintype == "audio":
                    attachment = MIMEAudio(content, _subtype=subtype)
                else:
                    attachment = MIMEBase(maintype, subtype)

                    attachment.set_payload(content)
                    encoders.encode_base64(attachment)

                attachment.add_header("Content-Disposition", "attachment", filename=attachment_filename.name)

                msg.attach(attachment)

            smtp.sendmail(from_addr, recipients, msg.as_string())

    # ----------------------------------------------------------------------
    def SendTestMessage(
        self,
        recipients: list[str],
        attachment_filenames: list[Path] | None = None,
    ) -> None:
        """Send a test message using the current profile."""

        self.SendMessage(
            recipients,
            f"SmtpMailer Verification ({datetime.datetime.now()})",  # noqa: DTZ005
            "This is a test message to ensure that the profile is working as expected.\n",
            attachment_filenames,
        )

    # ----------------------------------------------------------------------
    @classmethod
    def Load(
        cls,
        profile_name: str,
    ) -> "SmtpMailer":
        """Load a previously saved profile file."""

        data_filename = Path("~").expanduser() / (profile_name + cls.PROFILE_EXTENSION)

        if not data_filename.is_file():
            message = f"'{profile_name}' is not a recognized profile name."
            raise Exception(message)

        content = data_filename.read_bytes()

        if os.name == "nt":
            import win32crypt

            content = win32crypt.CryptUnprotectData(content, None, None, None, 0)[1]

        return cls(**json.loads(content.decode("utf-8")))

    # ----------------------------------------------------------------------
    @classmethod
    def EnumProfiles(cls) -> Generator[str, None, None]:
        """Enumerate all saved profiles."""

        for item in Path("~").expanduser().iterdir():
            if item.suffix == cls.PROFILE_EXTENSION:
                yield item.stem



import textwrap
import uuid

from pathlib import Path

import pytest

from dbrownell_Email.SmtpMailer import SmtpMailer


# ----------------------------------------------------------------------
def test_ToString():
    mailer = SmtpMailer("host", "username", "password", "from_name", "from_email", ssl=True, port=123)

    assert mailer.ToString() == textwrap.dedent(
        """\
        host        : host
        username    : username
        password    : ****
        from_name   : from_name
        from_email  : from_email
        ssl         : True
        port        : 123
        """,
    )


# ----------------------------------------------------------------------
def test_SaveAndLoad():
    profile_name = str(uuid.uuid4())
    profile_filename = Path("~").expanduser() / f"{profile_name}{SmtpMailer.PROFILE_EXTENSION}"

    mailer = _CreateSmtpMailer()

    # Save the profile
    assert not profile_filename.exists(), profile_filename
    mailer.Save(profile_name)
    assert profile_filename.exists(), profile_filename

    # Load the profile
    loaded_mailer = SmtpMailer.Load(profile_name)
    assert loaded_mailer == mailer

    # Remove the profile file
    profile_filename.unlink()


# ----------------------------------------------------------------------
def test_LoadError():
    with pytest.raises(Exception, match="'Does not exist' is not a recognized profile name."):
        SmtpMailer.Load("Does not exist")


# ----------------------------------------------------------------------
def test_EnumProfiles():
    # Create a couple of profiles
    mailer1 = _CreateSmtpMailer()
    mailer2 = _CreateSmtpMailer()

    mailer1.Save(mailer1.host)
    mailer2.Save(mailer2.host)

    # Find the profiles
    found_mailer1 = False
    found_mailer2 = False

    for profile in SmtpMailer.EnumProfiles():
        if profile == mailer1.host:
            assert found_mailer1 is False
            found_mailer1 = True
        elif profile == mailer2.host:
            assert found_mailer2 is False
            found_mailer2 = True

    assert found_mailer1 is True
    assert found_mailer2 is True

    (Path("~").expanduser() / f"{mailer1.host}{SmtpMailer.PROFILE_EXTENSION}").unlink()
    (Path("~").expanduser() / f"{mailer2.host}{SmtpMailer.PROFILE_EXTENSION}").unlink()


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _CreateSmtpMailer(
    host: str | None = None,
    username: str | None = None,
    password: str | None = None,
    from_name: str | None = None,
    from_email: str | None = None,
    ssl: bool | None = None,
    port: int | None = None,
) -> SmtpMailer:
    return SmtpMailer(
        host or str(uuid.uuid4()),
        username or str(uuid.uuid4()),
        password or str(uuid.uuid4()),
        from_name or str(uuid.uuid4()),
        from_email or str(uuid.uuid4()),
        ssl=ssl if ssl is not None else True,
        port=port,
    )



# noqa: D104

# Wheel names will be generated according to this value. Do not manually modify this value; instead
# update it according to committed changes by running this command from the root of the repository:
#
#   uv run python -m AutoGitSemVer.scripts.UpdatePythonVersion ./src/dbrownell_Email/__init__.py ./src
#
__version__ = "0.2.1"

from .SmtpMailer import SmtpMailer

__all__ = [
    "SmtpMailer",
]
